/* $Header: /fridge/cvs/libjcgi/cgi-env.c,v 1.10 2001/12/10 22:56:34 justins Exp $ */
/*
   
   JCGI - cgi-env.c           Copyright(c) 2001 Justin David Smith
   justins(at)chaos2.org      http://chaos2.org/
    
   Utilities for manipulating the CGI environment
    

   This program is free software; you can redistribute it and/or modify 
   it under the terms of the GNU General Public License as published by 
   the Free Software Foundation, version 2 of the License ONLY.

   This program is distributed in the hope that it will be useful, 
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation, 
   Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/
#include <assert.h>
#include <libjcgi.h>
#include <libjtree.h>


bintree *parse_query_args(const char *query_string) {
/* Parse out the query string's options.  The query string
   passed here must be the *unescaped* version of the query
   string. */

   stringa *arg;
   stringa *key;
   stringa *val;
   bintree *opt;
   int count;
   int i;

   /* Make sure the input is valid */
   if(query_string == NULL) return(NULL);
   
   /* Allocate a pair tree to store arguments in */
   opt = bintree_new(bintree_pair_lessthan, (destructorfn)pair_free);
   if(opt == NULL) return(NULL);
   
   /* Read the arguments */
   arg = strnewa(NULL);
   key = strnewa(NULL);
   val = strnewa(NULL);
   count = getnumargs(query_string, '&');
   for(i = 0; i < count; ++i) {
      /* Read the key name and the value */
      getarga(arg, query_string, i, '&');
      getargaa(key, arg, 0, '=');
      getargaa(val, arg, 1, '=');
      
      /* Unescape the key and values */
      unescapeCGIa(key);
      unescapeCGIa(val);
      cgi_check_input_chars_html(key);
      cgi_check_input_chars_html(val);
      
      /* Save the key/value pair to the tree */
      bintree_insert(opt, pair_alloc(stra(key), stra(val)));
   }
   
   /* Return the final options */
   strreleasea(&arg);
   strreleasea(&key);
   strreleasea(&val);
   return(opt);

}


static bool _patch_query_string(stringa *query_string) {
/* Patches up the input query string.  Some browsers/proxies
   do not unescape &blah; entities like they're supposed to
   when they appear in a URL; this function tries to clean 
   those occurrences up, although we have to break the std. 
   a bit to do it (ugh) */

   return(strreplacea(query_string, "&amp;", NULL) > 0);

}


void cgi_env_init(cgi_env *e) {
/* Create a new CGI environment using the environment
   variables for a baseline.  Be sure to set reasonable
   defaults where appropriate, and make sure the vars
   do not contain illegal characters!  */

   assert(e != NULL);
   
   /* Do this to avoid segfaults when running from command line. */
   setenv("QUERY_STRING",     "", false);
   setenv("REMOTE_ADDR",      "127.0.0.1", false);
   setenv("REMOTE_HOST",      getenv("REMOTE_ADDR"), false);
   setenv("HTTP_USER_AGENT",  "Targeted v2.0", false);
   setenv("HTTP_REFERER",     "-", false);
   setenv("SERVER_NAME",      "localhost", false);
   setenv("REQUEST_URI",      "/", false);
   
   /* Allocate input strings */
   e->query_string      = strnewa(getenv("QUERY_STRING"));
   e->remote_addr       = strnewa(getenv("REMOTE_ADDR"));
   e->remote_host       = strnewa(getenv("REMOTE_HOST"));
   e->http_user_agent   = strnewa(getenv("HTTP_USER_AGENT"));
   e->http_referer      = strnewa(getenv("HTTP_REFERER"));
   e->server_name       = strnewa(getenv("SERVER_NAME"));
   e->request_uri       = strnewa(getenv("REQUEST_URI"));
   
   /* Parse out the CGI options */
   e->query_edit = _patch_query_string(e->query_string);
   e->query_args = parse_query_args(stra(e->query_string));
   
   /* Unescape the CGI query string */
   unescapeCGIa(e->query_string);
   
   /* Sanity checking */
   cgi_check_input_chars_html(e->query_string);
   cgi_check_input_chars_html(e->remote_addr);
   cgi_check_input_chars_html(e->remote_host);
   cgi_check_input_chars_html(e->http_user_agent);
   cgi_check_input_chars_html(e->http_referer);
   cgi_check_input_chars_html(e->server_name);
   cgi_check_input_chars_html(e->request_uri);

}


cgi_env *cgi_env_new(void) {
/* Create a new CGI environment using the environment
   variables for a baseline.  Unlike above, this one
   allocates a new data structure.  */
   
   cgi_env *e;

   /* Attempt to allocate the CGI environment */
   e = (cgi_env *)malloc(sizeof(cgi_env));
   if(e == NULL) return(NULL);
   
   /* Initialise */
   cgi_env_init(e);
   return(e);

}


void cgi_env_free_entries(cgi_env *e) {
/* Released the environemnt */

   if(e == NULL) return;
   strreleasea(&e->query_string);
   strreleasea(&e->remote_addr);
   strreleasea(&e->remote_host);
   strreleasea(&e->http_user_agent);
   strreleasea(&e->http_referer);
   strreleasea(&e->server_name);
   strreleasea(&e->request_uri);
   bintree_release(&e->query_args);

}


void cgi_env_free(cgi_env *e) {
/* Released the environment */

   if(e == NULL) return;
   cgi_env_free_entries(e);
   free(e);

}


void cgi_env_release(cgi_env **e) {
/* Released the environemnt */

   if(e == NULL) return;
   cgi_env_free(*e);
   *e = NULL;

}


